<?php

namespace Drupal\content_fixtures\Loader;

use Drupal\content_fixtures\Exception\CircularReferenceException;
use Drupal\content_fixtures\Fixture\DependentFixtureInterface;
use Drupal\content_fixtures\Fixture\FixtureGroupInterface;
use Drupal\content_fixtures\Fixture\FixtureInterface;
use Drupal\content_fixtures\Fixture\OrderedFixtureInterface;
use Drupal\content_fixtures\Fixture\SharedFixtureInterface;
use Drupal\content_fixtures\Service\DatabaseCleanerInterface;
use Drupal\content_fixtures\Service\ReferenceRepositoryInterface;

/**
 * Class Loader
 *
 * @package Drupal\content_fixtures
 */
class Loader implements LoaderInterface {

  /**
   * Array of fixtures.
   *
   * @var FixtureInterface[]
   *
   * @see Loader::addGenerator
   */
  private $fixtures = [];

  /**
   * @var ReferenceRepositoryInterface
   */
  private $referenceRepository;

  /** @var DatabaseCleanerInterface */
  private $databaseCleaner;

  public function __construct(ReferenceRepositoryInterface $referenceRepository, DatabaseCleanerInterface $databaseCleaner) {
    $this->referenceRepository = $referenceRepository;
    $this->databaseCleaner = $databaseCleaner;
  }

  /**
   * @param FixtureInterface $fixture
   */
  public function addFixture(FixtureInterface $fixture) {

    if ($fixture instanceof SharedFixtureInterface) {
      $fixture->setReferenceRepository($this->referenceRepository);
    }

    if ($this->findObjectInArray(\get_class($fixture), $this->fixtures)) {
      throw new \RuntimeException('There was an attempt to add the same fixture twice to the loader.');
    }

    if ($fixture instanceof OrderedFixtureInterface && $fixture instanceof DependentFixtureInterface) {
      throw new \InvalidArgumentException('Fixture cannot implement both: OrderedFixtureInterface and DependentFixtureInterface .');
    }

    $this->fixtures[] = $fixture;
  }

  /**
   * @inheritdoc
   */
  public function getFixtures(array $groups = []) {

    if (empty($groups)) {
      return $this->sortFixtures($this->fixtures);
    }

    $fixtures = array_filter($this->fixtures, function(FixtureInterface $fixture) use ($groups) {
      return $fixture instanceof FixtureGroupInterface && !empty(array_intersect($groups, $fixture->getGroups()));
    });



    return $this->sortFixtures($fixtures);
  }

  /**
   * @inheritdoc
   */
  public function cleanup() {
    $this->databaseCleaner->run();
  }

  /**
   * Loads all fixtures.
   */
  public function loadFixtures() {
    $fixtures = $this->getFixtures();

    foreach ($fixtures as $fixture) {
      $fixture->load();
    }
  }

  /**
   * Orders fixtures by dependencies
   *
   * @param array $fixtures
   *
   * @return FixtureInterface[]
   */
  private function sortFixtures(array $fixtures) {

    /*
     * Sort by numbers.
     */
    usort($fixtures, function ($a, $b) {
      if ($a instanceof OrderedFixtureInterface && $b instanceof OrderedFixtureInterface) {
        if ($a->getOrder() === $b->getOrder()) {
          return 0;
        }
        return $a->getOrder() < $b->getOrder() ? -1 : 1;
      }
      if ($a instanceof OrderedFixtureInterface) {
        return $a->getOrder() === 0 ? 0 : 1;
      }
      if ($b instanceof OrderedFixtureInterface) {
        return $b->getOrder() === 0 ? 0 : -1;
      }
      return 0;
    });

    /*
     * Sort by dependencies.
     */

    $sequenceForClasses = [];

    // First we determine which classes have dependencies and which don't
    foreach ($fixtures as $fixture) {
      $fixtureClass = get_class($fixture);

      if ($fixture instanceof DependentFixtureInterface) {
        $dependenciesClasses = $fixture->getDependencies();

        $this->validateDependencies($dependenciesClasses, $fixtures);

        if (!is_array($dependenciesClasses) || empty($dependenciesClasses)) {
          throw new \InvalidArgumentException(sprintf('Method "%s" in class "%s" must return an array of classes which are dependencies for the fixture, and it must be NOT empty.', 'getDependencies', $fixtureClass));
        }

        if (in_array($fixtureClass, $dependenciesClasses, TRUE)) {
          throw new \InvalidArgumentException(sprintf('Class "%s" can\'t have itself as a dependency', $fixtureClass));
        }

        // We mark this class as unsequenced
        $sequenceForClasses[$fixtureClass] = -1;
      } else {
        // This class has no dependencies, so we assign 0
        $sequenceForClasses[$fixtureClass] = 0;
      }
    }

    // Now we order fixtures by sequence
    $sequence = 1;
    $lastCount = -1;

    while (($count = count($unsequencedClasses = $this->getUnsequencedClasses($sequenceForClasses))) > 0 && $count !== $lastCount) {
      foreach ($unsequencedClasses as $key => $class) {
        $fixture = $this->findObjectInArray($class, $fixtures);
        $dependencies = $fixture->getDependencies();
        $unsequencedDependencies = $this->getUnsequencedClasses($sequenceForClasses, $dependencies);

        if (count($unsequencedDependencies) === 0) {
          $sequenceForClasses[$class] = $sequence++;
        }
      }

      $lastCount = $count;
    }

    $orderedFixtures = [];

    // If there're fixtures unsequenced left and they couldn't be sequenced,
    // it means we have a circular dependency.
    if ($count > 0) {
      throw new CircularReferenceException(sprintf('Classes "%s" have produced a CircularDependencyException.', implode(',', $unsequencedClasses)));
    }

    // We order the classes by sequence
    asort($sequenceForClasses);

    foreach ($sequenceForClasses as $class => $sequence) {

      $orderedFixtures[] = $this->findObjectInArray($class, $fixtures);
    }

    return $orderedFixtures;
  }

  private function getUnsequencedClasses($sequences, $classes = NULL) {
      $unsequencedClasses = [];

      if (NULL === $classes) {
          $classes = array_keys($sequences);
      }

      foreach ($classes as $class) {
          if ($sequences[$class] === -1) {
              $unsequencedClasses[] = $class;
          }
      }

      return $unsequencedClasses;
  }

  private function validateDependencies(array $dependenciesClasses, array $fixtures) {

      $loadedFixtureClasses = [];
      foreach ($fixtures as $fixture) {
        $loadedFixtureClasses[] = \get_class($fixture);
      }

      foreach ($dependenciesClasses as $class) {
          if (!in_array($class, $loadedFixtureClasses, TRUE)) {
              throw new \RuntimeException(sprintf('Fixture "%s" was declared as a dependency, but it should be added in fixture loader first.', $class));
          }
      }

      return TRUE;
  }

  /**
   * @param $class
   * @param array $array
   *
   * @return FixtureInterface|null
   */
  private function findObjectInArray($class, array $array) {

    foreach ($array as $element) {
      if (\get_class($element) === $class) {
        return $element;
      }
    }

    return NULL;
  }

}
